'use strict';

const _ = require('lodash');
const Promise = require('bluebird');
const Deferrari = require('deferrari');

const State = require('./state');
const Store = require('../../persistence');

const STATE = {
  INITIALIZED: 'initialized'
};

// This establishes a private namespace.
const namespace = new WeakMap();
function p(object) {
  if (!namespace.has(object)) namespace.set(object, {});
  return namespace.get(object);
}



/**
 *
 */
module.exports = class LambdaInvokee {
  constructor(config) {
    config = config || {};

    p(this).options = {
      log: {
        updates: _.get(config, 'options.log.updates', true),
        errors: _.get(config, 'options.log.errors', true)
      },
      store: {
        refreshOnUpdate: _.get(config, 'options.store.refreshOnUpdate', true),
        refreshOnError: _.get(config, 'options.store.refreshOnError', true)
      }
    };

    p(this).log = {
      notice: _.get(config, 'log.notice', console.log),
      error: _.get(config, 'log.error', console.log)
    };

    p(this).store = new Store(config.store);

    p(this).deferrari = new Deferrari({Promise});
  }

  /**
   * Returns AWS Lambda function.
   */
  getLambdaFunction(fn, config) {
    const self = this;
    config = config || {};
    return (payload, context, callback) => {
      _.set(config, 'state.context', context);
      const state = new State(config.state);

      
      let lambdaIdentifier = context.awsRequestId;
      const fnConfig = {
        setLambdaIdentifier: _lambdaIdentifier => lambdaIdentifier = _lambdaIdentifier
      };

      // Ensure that lambdaIdentifier is set. In timeout to remove from execution
      // flow until caller has opportunity to set value.
      setTimeout(() => {
        if (!lambdaIdentifier) lambdaIdentifier = context.awsRequestId;
        state.setMeta({lambdaIdentifier});
        p(this).deferrari.resolve(STATE.INITIALIZED);
      });

      const eventHandlers = {
        onUpdate: (data, config) => onUpdate(data, config),
        onError: (error, config) => onError(error, config)
      };

      return Promise.resolve(fn(payload, fnConfig, eventHandlers))
      .then(result => {
        onUpdate(result, {final: true});
        callback(null, result);
      })
      .catch(err => {
        onError(err, {final: true});
        callback(err);
      });

      function onUpdate(data, config) {
        // Defer so that client has had opportunity to set the lambdaIdentifier.
        return p(self).deferrari.deferUntil(STATE.INITIALIZED)
        .then(() => {
          state.update(data, config);

          const stateOutput = state.output();

          const options = p(self).options;
          if (options.log.updates) p(self).log.notice(stateOutput);
          if (options.store.refreshOnUpdate) {
            p(self).store.upsert(lambdaIdentifier, stateOutput);
          }
        });
      }

      function onError(err, config) {
        // Defer so that client has had opportunity to set the lambdaIdentifier.
        return p(self).deferrari.deferUntil(STATE.INITIALIZED)
        .then(() => {
          state.error(err, config);

          const options = p(self).options;
          if (options.log.errors) p(self).log.error(err && err.stack);
          if (options.store.refreshOnError) {
            const stateOutput = state.output();
            p(self).store.upsert(lambdaIdentifier, stateOutput);
          }
        });
      }
    };
  }
}
