'use strict';

const _ = require('lodash');
const Promise = require('bluebird');


// This establishes a private namespace.
const namespace = new WeakMap();
function p(object) {
  if (!namespace.has(object)) namespace.set(object, {});
  return namespace.get(object);
}



/**
 *
 */
module.exports = class LambdaState {
  constructor(config) {
    config = config || {};

    p(this).startedAt = Date.now();

    p(this).context = config.context;
    if (!p(this).context) throw new Error('LambdaState must be provided AWS Lambda context.');

    p(this).updates = [];
    p(this).updatedCount = 0;
    p(this).lastUpdated = Date.now();

    if (config.meta) this.setMeta(config.meta);
  }

  /**
   * Update the state, optionally include update data. If included, will be
   * added to state.
   */
  update(data, config) {
    config = config || {};
    const timestamp = p(this).lastUpdated = Date.now();
    p(this).updatedCount++;

    if (data != undefined) {
      p(this).updates.push({timestamp, data});
    }

    if (config.final) {
      if (p(this).resolved) throw new Error('Cannot set final lambda state to resolved if already resolved.');
      if (p(this).rejected) throw new Error('Cannot set final lambda state to resolved if already rejected.');
      p(this).resolved = true;
    }
  }

  error(err, config) {
    config = config || {};
    const timestamp = p(this).lastUpdated = Date.now();
    p(this).updatedCount++;

    if (err != undefined) {
      p(this).updates.push({timestamp, err: err && err.stack});
    }

    if (config.final) {
      if (p(this).resolved) throw new Error('Cannot set final lambda state to rejected if already resolved.');
      if (p(this).rejected) throw new Error('Cannot set final lambda state to rejected if already rejected.');
      p(this).rejected = true;
    }
  }

  /**
   * Output Lambda State as JSON.
   */
  output() {
    const now = Date.now();
    const context = p(this).context;

    const output = {
      time: {
        startedAt: p(this).startedAt,
        now: now,
        currentRunningTime: now - p(this).startedAt,
        remainingExcecutionTime: context.getRemainingTimeInMillis()
      },
      awsRequestId: context.awsRequestId,
      lambda: {
        name: context.functionName,
        arn: context.invokedFunctionArn,
        version: context.functionVersion,
        memoryLimit: context.memoryLimitInMB
      },
      updates: {
        count: p(this).updatedCount,
        lastUpdated: p(this).lastUpdated,
        data: p(this).updates
      }
    };

    if (p(this).meta) output.meta = p(this).meta;

    return output;
  }

  /**
   * Set meta property on state with any arbitrary data.
   */
  setMeta(meta) {
    p(this).meta = meta;
  } 
}
