'use strict';

const Promise = require('bluebird');
const AWS = require('aws-sdk');
const Deferrari = require('deferrari');

const STATE = {
  BUCKET_CREATED: 'bucket_created'
};


// This establishes a private namespace.
const namespace = new WeakMap();
function p(object) {
  if (!namespace.has(object)) namespace.set(object, {});
  return namespace.get(object);
}



/**
 * S3 Store.
 */
module.exports = class Store {
  constructor(config) {
    config = config || {};

    p(this).bucket = config.bucket;
    if (!config.bucket) throw new Error('S3 store must be configured with bucket.');

    p(this).S3 = new AWS.S3({
      region: config.region,
      credentials: new AWS.Credentials({
        accessKeyId: config.accessKeyId,
        secretAccessKey: config.secretAccessKey
      })
    });

    p(this).deferrari = new Deferrari({Promise});
  }

  /**
   * Initialize S3 Store.
   */
  initialize() {
    // Create buckets.
    return createBuckets(p(this).S3, p(this).bucket)
    .tap(response => p(this).deferrari.resolve(STATE.BUCKET_CREATED, response))
  }

  /**
   * Get value at key.
   */
  get(key) {
    return p(this).deferrari.deferUntil(STATE.BUCKET_CREATED)
    .then(() => {
      return Promise.resolve(p(this).S3.getObject({
        Bucket: p(this).bucket,
        Key: key
      }).promise())
      .then(response => {
        return JSON.parse(response.Body.toString('utf-8'));
      });
    });
  }

  /**
   * Upsert value at key.
   */
  upsert(key, value) {
    return p(this).deferrari.deferUntil(STATE.BUCKET_CREATED)
    .then(() => {
      return Promise.resolve(p(this).S3.putObject({
        Bucket: p(this).bucket,
        Key: key,
        Body: JSON.stringify(value)
      }).promise());
    });
  }
}

function createBuckets(s3Client, buckets) {
  buckets = buckets.replace(/\/+$/, '').split('/');
  const current = [buckets.shift()];
  
  return createBucket();

  function createBucket() {
    const remaining = buckets.length;
    return Promise.resolve(s3Client.createBucket({
      Bucket: current.join('/') + (remaining ? '/' : '')
    }).promise())
    .catch(err => {
      if (err && err.name === 'BucketAlreadyOwnedByYou') return;
      return Promise.reject(err);
    })
    .then(() => {
      if (remaining) {
        current.push(buckets.shift());
        return createBucket();
      }
    });
  }
}

